##################################################
# Optimization of a distillation column with 
# parameters Energy price, pV, and feed rate, F.
##################################################

#---------------------------------
# Setting parameters as variables 
#---------------------------------

var F ;
var zF ;
var qF ;
var pV ;

#---------------------------
# PARAMETERS
#---------------------------

param NT, := 41 ; 	        # number of stages including reboiler
param NF, := 21 ; 	# location of of feed stage (counted from the bottom)

### P-Controllers for control of reboiler and condenser hold up

param KcB, := 10 ;		# controller gain reboiler
param KcD, := 10 ;		# controller gain condenser
param MDs, := 0.5 ;		# nominal holdup condenser
param MBs, := 0.5 ;		# nominal holdup reboiler
param Ds, := 0.5  ;		# nominal flow condenser
param Bs, := 0.5 ;		# nominal flow reboiler

param alpha, := 1.5 ;		# relative volatility
param M0, := 0.5 ; 		#nominal liquid holdups

### Data for linearized liquid flow dynamics (does not apply to reboiler or condenser)

param tau1, := 0.063 ;		# time constant for lquid dynamics (min)
param F0, := 1 ; 		# nominal feed rate (kmol/min)
param qF0, := 1 ;		# nominal fraction of liquid 
param L0, := 2.70629 ;		# nominal reflux flow (from steady-state data)
param L0b, >= 0 ;		# nominal liquid flow below flow (kmol/min)
param lambda, := 0 ;		# effect of vapor flow on liquid flow ("K2-effect")
param V0, := 3.20629 ;		# nominal vapor flow
param V0t, >= 0 ;		# nominal vapor flow

### Define prices

param pF, := 1 ;		# feed price ($) 
param pB, := 1 ;		# bottoms price ($)
param pD, := 2 ;		# price distillate ($)


#---------------------------------
# VARAIBLES
#---------------------------------
var LT >= 0.1 <= 10 := 2.76969 ;
var VB >= 0.1 <= 10 := 3.39585 ; 

var L {2..NT} >= 0 ;
var V {1..NT-1} >= 0 ;

var D >= 0 ;			# distillate flow
var B >= 0 ;			# bottoms flow

var MB >= 0 ;    		# actual reboiler holdup
var MD >= 0 ;			# actual condenser holdup

var x {1..2*NT} >= 0 ;
var y {1..NT-1} >= 0 ;

var pD2 ;			# new distillate price
 
#------------------------------------------------------------------
# Initial values for stages/holdup and distillate and bottoms flow
#------------------------------------------------------------------
for {i in 1..2*NT}
let x[i] := 0.5 ;		# initial values for composition in the stages

let L0b := L0 + qF0*F0 ;
let V0t := V0 + (1 - qF0)*F0 ; 

#---------------
# CONSTRAINTS
#---------------


### Control liquid level in reboiler and condenser
const1: 
		MB = x[NT+1] ;
const2:
		MD = x[2*NT] ;
const3:
		B = Bs + (MB - MBs)*KcB ;
const4:
 		D = Ds + (MD - MDs)*KcD ;

### Model
const5 {i in 1..NT-1}:
		y[i] = alpha*x[i]/(1 + (alpha-1)*x[i]) ;					# vapor-liquid equilibria
		
const6 {i in 1..NF-1}:										
		V[i] = VB ;													# vapor flow rectifier
	
const7 {i in NF..NT-1}:										
		V[i] = VB + (1-qF)*F ;										# vapor flow stripper

const8 {i in 2..NF}:										
		L[i] = L0b + (x[i+NT] - M0)/tau1 + lambda*(V[i-1] - V0) ;	# liquid flow rectifier

const9 {i in NF+1..NT-1}:
		L[i] = L0  + (x[i+NT] - M0)/tau1 + lambda*(V[i-1] - V0t) ;	# liquid flow stripper

const10: 
		L[NT] = LT ;												# liquid flow condenser

const11 {i in 2..NF-1}: 
		L[i+1] - L[i] + V[i-1] - V[i] = 0 ; 							# dMdt rectifier

const12 {i in 2..NF-1}: 
		L[i+1]*x[i+1] - L[i]*x[i] + V[i-1]*y[i-1] - V[i]*y[i] = 0 ;		# dMxdt rectifier

const13 {i in NF+1..NT-1}:	
		 L[i+1] - L[i] + V[i-1] - V[i] = 0 ; 							# dMdt stripper

const14 {i in NF+1..NT-1}:		
		L[i+1]*x[i+1] - L[i]*x[i] + V[i-1]*y[i-1] - V[i]*y[i] = 0 ; 	# dMxdt stripper

const15:
		L[2] - V[1] - B = 0 ;											# dMdt reboiler

const16:
		L[2]*x[2] - V[1]*y[1] - B*x[1] = 0 ;							# dMxdt reboiler

const17:
		L[NF+1] - L[NF] + V[NF-1] - V[NF]  + F = 0 ;  								# dMdt feed

const18: 
		L[NF+1]*x[NF+1] - L[NF]*x[NF] + V[NF-1]*y[NF-1] - V[NF]*y[NF] + F*zF = 0 ;	# dMxdt feed

const19:
		V[NT-1] - LT - D = 0 ;											# dMdt condenser
	
const20:
		V[NT-1]*y[NT-1] - LT*x[NT] - D*x[NT] = 0 ; 						# dMxdt condenser


### Inequality constraints
const21: 
 		1 - x[1] >= 0.99 ;						# fraction of heavy comp in distillate
 
const22: 

 		x[NT] >= 0.95 ; 						# fraction of light comp in bottoms
 
const23: 
 		VB <= 4.008 ;							# max vapor flow


### Case 1b) pert
const24:
		pD2 = pD*x[NT] ;						# distillate price proportional to purity xD

### Initial constraints for parameters
constF: F = nominal_F ;
constpV: pV = nominal_pV ;
constzF: zF = nominal_zF ;
constqF: qF = nominal_qF ;

#--------------------
# OBJECTIVE FUNCTION
#--------------------
minimize cost: pF*F + pV*VB - pB*B - pD2*D ; 

